# -*- coding: utf-8 -*-
"""
Created on Thu May 10 11:36:05 2018

@author: Michael Stadlhofer 01530415
"""

from timeit import default_timer as timer
import numpy as np
from scipy import special, integrate, misc

orbital_dict = {
        (1,0,0): r"$1s$",
        (2,0,0): r"$2s$",
        (2,1,0): r"$2p_z$",
        (2,1,1): r"$2p_x$",
        (2,1,-1): r"$2p_y$",
        (3, 0,0): r"$3s$",
        (3, 1, 0): r"$3p_z$",
        (3, 1, 1): r"$3p_x$",
        (3, 1, -1): r"$3p_y$",
        (3, 2, 0): r"$3d_{zz}$",
        (3, 2, 1): r"$3d_{zx}",
        (3, 2, -1): r"$3d_{zy}$", 
        (3, 2, 2): r"$3d_{xy}$",
        (3, 2, -2): r"$3d_{x^2-y^2}$",
        (4,0,0): "$4s$"}

slater_dict = {#(Z, n)
        (1, 1): 1,
        (2, 1): 1.7,
        (3, 1): 2.7,
        (4, 1): 3.7,
        (5, 1): 4.7,
        (6, 1): 5.7,
        (7, 1): 6.7,
        (8, 1): 7.7,
        (9, 1): 8.7,
        (10, 1): 9.7,
        (3, 2): 1.3,
        (4, 2): 1.95,
        (5, 2): 2.6,
        (6, 2): 3.25,
        (7, 2): 3.9,
        (8, 2): 4.55,
        (9, 2): 5.2,
        (10, 2): 5.85    
        }


#constants
#eV in J
eV=1.602*10**(-19)


#hbar in Js
hbar = 1.05457182e-34

#epsilon0 in C^2 / Nm^2
epsilon0 = 8.854e-12

# elementary charge e in  C
q = 1.602176634e-19

# electron mass me in kg
me = 9.109e-31

#bohr radius in m
a0 = 4*np.pi*epsilon0*hbar**2*(1/(me*q**2))

#degree
degree = np.pi/180



# dimensions of the molecule
# used for determining the limits of the 2d and 3d plots
def mol_dim(mol):

    atompos = []
    for _ in mol.atoms:
       atompos.append(_.r0)
       temp = np.asarray(atompos).transpose()
       xpos, ypos, zpos = temp[0], temp[1], temp[2]
       xdist = max(xpos) - min(xpos)
       ydist = max(ypos) - min(ypos)
       zdist = max(zpos) - min(zpos)

    chardist = max([xdist, ydist, zdist])

    xmin = min(xpos) - chardist/2.0
    xmax = max(xpos) + chardist/2.0
    ymin= min(ypos) - chardist/2.0
    ymax = max(ypos) + chardist/2.0
    zmin = min(zpos) - chardist/2.0
    zmax = max(zpos) + chardist/2.0
    
    return chardist, [xmin, xmax, ymin, ymax, zmin, zmax]


# numerical derivative
def D(f, n, dx = a0/100):
    return lambda y: misc.derivative(f, y, dx=dx,n=n)

    
# atomic orbital class          
#   provides functions for the orbital wavefunction (R, Y, psi)
#   the coulomb potential (V, V_vec)
#   aswell as kinetic energy T and hamiltonian H applied to the wavefunction
class ao:

    def __init__(self, Z, n, l, m, r0, slater = False):

        # atomic number
        # use slater rules for the atomic charge
        if n<=2 and Z<11 and slater == True:
            self.Z = slater_dict[Z,n]
        else:
            self.Z = Z
            
        #quantum numbers
        self.n = n      # principal 
        self.l = l      # angular momentum 
        self.m = m      # magnetic
        #check if r0 is an array, if not change it to an array
        if type(r0)==list:
            r0 = np.asarray(r0)
        self.r0 = r0    # position of the atom
        
        
     
            
            
        # energy of the atomic orbital
        self.E = -self.Z**2*me*q**4/(32*np.pi**2*epsilon0**2*hbar**2*n**2)/eV

        
    #radial part of the wavefunction
    def R(self, r):
        rho = 2*self.Z*r/(self.n*a0)
        const = np.sqrt((2*self.Z/(self.n*a0))**3*np.math.factorial(self.n-self.l-1)/(np.math.factorial(self.n+self.l)*2*self.n))
        return const*np.exp(-rho/2)*rho**self.l*special.eval_genlaguerre(self.n - self.l - 1, 2*self.l + 1, rho)
    
    # spherical harmonic
    def Y(self, theta, phi):
        return special.sph_harm(self.m, self.l, phi, theta)
    

    # potential caused by the atom at a position in R3
    def V_vec(self, x,y,z, r0= np.asarray([0,0,0])):
        # r0 = position of the atom
        return q**2*self.Z/(4*np.pi*epsilon0*np.sqrt((x-r0[0])**2 + (y-r0[1])**2 + (z-r0[2])**2))
    
    
    # wavefunction psi dependent on r, theta and phi and the position of the atom r0
    def psi(self, r, theta, phi, r0 = np.asarray([0,0,0])):
        if np.linalg.norm(r0)==0:
            return self.R(r)*self.Y(theta, phi)
        else:
            x = r*np.sin(theta)*np.cos(phi)  - r0[0]
            y = r*np.sin(theta)*np.sin(phi) - r0[1]
            z = r*np.cos(theta) - r0[2]
    
            r = np.sqrt(x**2 + y**2 + z**2)
            theta = np.arccos(z/r)
            phi = np.arctan2(y,x)
    
            return self.R(r)*self.Y(theta, phi)
        

    ### the following functions are not used in the main calculations
    # wavefunction psi dependent x, y and z
    def psi_xyz(self, x,y,z, r0 = np.asarray([0,0,0])):
        x -= r0[0]
        y -= r0[1]
        z -= r0[2]
        
        r = np.sqrt(x**2 + y**2 + z**2)
        return self.R(r)*self.Y(np.arccos(z/r), np.arctan2(y,x))
        
    # potential caused by the atom at a distance r from the atom
    def V(self, r):
        return q**2*self.Z/(4*np.pi*epsilon0*r)
    
    # kinetic energy term of the hamiltonian applied to the wavefunction
    def T(self, r, theta, phi):
        return -hbar**2/(2*me)*((self.Y(theta, phi)/r)*(2*D(self.R, 1)(r) + r*D(self.R, 2)(r)) - self.l*(self.l+1)/(r**2)*self.R(r)*self.Y(theta, phi))
    
    # hamiltonian applied to the wavefunction
    def H(self, r, theta, phi):
           return self.T(r,theta, phi) - self.V(r)*self.R(r)*self.Y(theta, phi)
   
    
    
# atom class 
#   -creates and stores orbitals for given quantumnumbers,
#       atomic number  Z and position of the atom r0
class atom:
    def __init__(self,quantumnumbers,Z, r0, symbol = " ", slater = False):
 
            
        # stores parameters that were given
        self.qn = quantumnumbers
        self.r0 = r0
        self.Z = Z
        self.symbol = symbol
        
        # list of all orbitals 
        self.orbitals = []
        for item in quantumnumbers:
            n, l, m = item
            self.orbitals.append(ao(Z, n, l, m, r0, slater))
            
        
        

# molecule class
#  -provides the braket integral function that is used to 
#       calculate the overlap and the hamiltonian matrix
#  -stores a list of the atoms 
#       and orbitals involved
#  -provides a function for the coulomb potential
#       of the sum of all atoms in the molecule (Vmo, Vmo_xyz) 
#  -can save information about the molecule and 
#       the solution to the roothaan equation to .txt files
#       which can be read by the visualisation programs
class molecule:
    def __init__(self, atoms = None, name = None):   
        self.name = name                    # name of the molecule
        self.atoms = atoms                  # atoms of the molecule
        self.orbitallist = []               # list of orbitals
        self.atomlist = []                  # list of atoms corresponding to
                                            #   the orbitals in the orbitallist
        
        self.orbitalnames = []              #  names of orbitals 
        
        # filling the orbitallist and the atomlist       
        if atoms != None:                         
            for atom in self.atoms:
                for orbital in atom.orbitals:
                    self.orbitallist.append(orbital)
                    self.atomlist.append(atom)
                    
        for _ in range(len(self.orbitallist)):
            self.orbitalnames.append(self.atomlist[_].symbol + "_" + list_txt([self.orbitallist[_].n, self.orbitallist[_].l,self.orbitallist[_].m],""))
        # variable for storing if the roothaan equations have been solved or not
        self.solved = False    
        
        
    # braket integral function
    def braket(self, ao1, ao2, O=1, rmax = None):
        ##parameters:
        # ao1 ... first atomic orbital (bra) centered around atom 1
        # ao2 ... second atomic orbital (ket) centered around atom 2
        # O ... Operator
        # ->  <ao1| O |ao2>
        
        
        # limits of the integration variable r
        rmin = 0
        if rmax == None: rmax = 60*a0
        # rmax = 50a0 yields the best results for H2O
        
        # position vector that connects the atom of ao2 to the atom of ao1
        d = ao1.r0 - ao2.r0 
        
        ## operator cases
        
         # O = H (hamilton operator of atom 2)
        # this can be used to calculate the energy of the orbital around atom 2
        # other than checking if the calculation yields the correct result (by comparison to self.E)
        #  serves no purpose in the main calculations of this program
        if O=="H":
            integrand = lambda r, theta, phi: np.conjugate(ao1.psi(r, theta, phi, d))*ao2.H(r, theta, phi)*r**2*np.sin(theta)
            return integrate.nquad(integrand,[[rmin, rmax], [0,np.pi], [0, 2*np.pi]])[0]/eV
        
        
        # atom around which the second orbital ao2 is centered
        atom2 = self.atomlist[self.orbitallist.index(ao2)]
        
        # indices of the involved orbitals
        i, j = [self.orbitallist.index(ao1), self.orbitallist.index(ao2)]
        
        # O = 1 (no operator) braket: <ao1|ao2>
        if O==1:
            # if the orbitals ao1 and ao2 belong to the same atom,
            if np.linalg.norm(d)==0:
                # compare the quantum numbers   
                if ao1.n == ao2.n and ao1.l == ao2.l and ao1.m==ao2.m:
                    #  orbitals around the same atom are orthogonal
                    #  -> if quantum numbers are  identical:
                    return 1
                else:
                    return 0
            # if the orbitals ao1 and ao2 are centered  around different atoms,
            else:
                # check if the overlap integral has already been calculated 
                    # (this is possible because <ao1|ao2> = <ao2|ao1>)
                if self.S[j,i] !=0:
                    return self.S[j,i]
                # otherwise evaluate the integral
                else:
                    integrand = lambda r, theta, phi: np.real(np.conjugate(ao1.psi(r, theta, phi, d))*ao2.psi(r, theta, phi))*r**2*np.sin(theta)
                    return integrate.nquad(integrand,[[rmin, rmax], [0,np.pi], [0, 2*np.pi]])[0]
        
       
        # O = Hmo (molecula hamiltonian)
        if O=="Hmo":
             # for an explanation of this calculation go to: https://lampx.tugraz.at/~hadley/ss1/molecules/hueckel/mo_h2_plus.php
            integrand = lambda r, theta, phi: np.real(np.conjugate(ao1.psi(r, theta, phi, d))*(self.Vmo(r, theta, phi,  origin = ao2.r0, exclude = atom2)*ao2.R(r)*ao2.Y(theta,phi)))*r**2*np.sin(theta)
            sigma_integral = integrate.nquad(integrand,[[rmin, rmax], [0,np.pi], [0, 2*np.pi]])[0]
            return ao2.E*self.S[i,j],  sigma_integral/eV
        
            ## this is the direct way of calculating this integral:
            ## it is however  slower and not as consistent
            #integrand = lambda r, theta, phi: np.conjugate(ao1.psi(r, theta, phi, d))*(ao2.T(r,theta,phi) - self.Vmo(r, theta, phi, origin = ao2.r0)*ao2.R(r)*ao2.Y(theta,phi))*r**2*np.sin(theta)
            #return integrate.nquad(integrand,[[rmin, rmax], [0,np.pi], [0, 2*np.pi]])[0], 0



 
    # this function solves the roothaan equations for the given atoms and orbitals
    def solve(self):
        # starttime of calculation
        t_init_0 = timer()
        
        # calculating the elements of the overlap matrix self.S
        self.S = np.zeros((len(self.orbitallist), len(self.orbitallist)))
        print("\n Calculating overlap matrix")
        for i in range(len(self.orbitallist)):
            for j in range(len(self.orbitallist)):
                t0 = timer() #starttime for the integral
                self.S[i,j] = self.braket(self.orbitallist[i], self.orbitallist[j])
                t1 = timer() #endtime for the integral
                print("  S" + str(1+ i)+ str(1+j) + " (" + str(round(t1 - t0,1)) + " s)")
        print("\n")
        
        # calculating the elements of the hamiltonian matrix self.H
        self.H = np.zeros(np.shape(self.S))
        
        # potential matrix sigma
        self.sigma  = np.zeros(np.shape(self.S))
        print(" Calculating hamilton matrix")
        for i in range(len(self.orbitallist)):
            for j in range(len(self.orbitallist)):
                t0 = timer() #starttime for the integral
                SE, sigma_val = self.braket(self.orbitallist[i], self.orbitallist[j], O="Hmo")
                self.H[i,j] = SE - sigma_val
                self.sigma[i,j] = sigma_val
                t1 = timer() #endtime for the integral
                print("  H" + str(1 + i)+ str(1+ j) + " (" + str(round(t1 - t0,1)) + " s)")
        print("\n")
        
        # calculating the inverse of the overlap matrix
        self.S_inv = np.linalg.inv(self.S)
        
        # solving the eigenvalue problem  S_inv H c = E c
        self.sol = np.linalg.eig(np.dot(self.S_inv, self.H))
        
        # sorting the solution by the size of the eigenvalues (energies)
        self.sol= [(self.sol[0][i], self.sol[1][:,i]) for i in range(len(self.sol[0]))]
        self.sol.sort(key = lambda x: -abs(x[0]))
        
        # end time of calculation
        t_init_1 = timer()
        print("Solving the roothaan equations for the " + self.name + " molecule took " + str(round(t_init_1 - t_init_0, 2)) + " seconds. \n \n")
        
        # equations have been solved
        self.solved = True
    
    # molecular potential (polar coordinates)
    #   -sum of the coulomb potentials of all the atoms in the molecule
    #   -depends on the polar coordinates r theta and phi around a specified origin
    def Vmo(self, r, theta ,phi, origin = [0,0,0], exclude = None):
        
        # in order to calculate the hamilton matrix elements more easily 
        # it is necessary to be able to exclude one atom from the sum
        # exclude: index of the atom to be excluded in the atomlist
        
        atms = self.atoms[:]
        if exclude != None:
            atms.remove(exclude) # here one atom is removed
            
        x = r*np.sin(theta)*np.cos(phi) + origin[0]
        y = r*np.sin(theta)*np.sin(phi) + origin[1]
        z = r*np.cos(theta) + origin[2]
       
        return sum([atom.orbitals[0].V_vec(x, y, z, atom.r0 ) for atom in atms])
    
    # molecular potential (carthesian coordinates)
    # with the option to include only one atom
    #  this function is used for plotting the 3d potential and for visualizing 
    #  the positions of the atoms
    def Vmo_xyz(self, x, y ,z, include = None):
        atms = self.atoms[:]
        if include != None:
            atms = [atms[include]]
        return sum([atom.orbitals[0].V_vec(x, y, z, atom.r0 ) for atom in atms])
    
    # function for saving the solution of the roothaan equations
    #   and the structure of the molecule to .txt files
    def save_molecule_data(self):
        if self.solved == True:
            with open(self.name + "_solution.txt", "w") as file:
                    for solution in self.sol:
                        energy = str(round(solution[0], 2))+"\n"
                        file.write(energy)
                        eigenvector = list_txt(solution[1])
                        file.write(eigenvector+"\n")
        
        with open(self.name + "_construction.txt", "w") as file:
            for Atom in self.atoms:
                file.write(Atom.symbol + "\n")
                file.write(list_txt(Atom.r0)+"\n")
                file.write(str(Atom.Z)+"\n")
                quantumnumbers = list_txt([list_txt(x) for x in Atom.qn])
                file.write(quantumnumbers+"\n")
                
                
   
             
### functions for plotting the molecular orbitals
 ###############################################    
           
# for 3d isosurface plots:
# wavefunction centered around r0 at a point in space (x, y, z) multiplied by a weight            
def psi_cart(orbital, weight, x,y,z, r0):
        r = np.sqrt((x - r0[0])**2 + (y-r0[1])**2 + (z-r0[2])**2)
        theta = np.arccos(z/r)
        phi = np.arctan2(y,x)
        
        return weight*orbital.R(r)*orbital.Y(theta, phi)
      
# sum of a linear combination of atomic orbitals
def aoSum(orbitals, weights):
    # returns a function of x, y and z
    return lambda x, y, z: sum([psi_cart(orbitals[i], weights[i], x,y,z, orbitals[i].r0) for i in range(len(orbitals))])


# for 2d plots:
# sum of a linear combination of atomic orbitals
def aoSum_xy(orbitals, weights):
    # returns a function of x and y
    return lambda x, y: sum([psi_cart(orbitals[i], weights[i], x,y,0, orbitals[i].r0) for i in range(len(orbitals))])   


# potential in the x, y plane 
def V_xy(mol):
    return lambda x, y: mol.Vmo(np.sqrt(x**2 + y**2),np.pi/2, np.arctan2(y,x), origin=[0,0,0])





### functions for dealing with txt files
 #######################################
   
# turns values of a list into whitespace separated string
def list_txt(l, seperator = " "):
    return seperator.join([str(value) for value in l])


# functions for loading the files created by molecule.save_molecule_data() 
def load_construction(name):
    File = []
    with open(name) as inputfile:
        for line in inputfile:
            File.append(line)
        
    loaded = []
    for i in range(len(File)//4):
        symbol = File[i*4].strip()
        r0 = [float(x) for x in File[i*4 + 1].split()]
        z = int(File[i*4 + 2])
        qn = [int(x) for x in File[i*4+3].split()]
        qn_list = []
        for j in range(len(qn)//3):
            qn_list.append([qn[j*3], qn[j*3 + 1], qn[j*3 + 2]])
        loaded.append([qn_list, z,r0, symbol])      
    return loaded

def load_solution(name):
    File = []
    with open(name) as inputfile:
        for line in inputfile:
            File.append(line)
            
    loaded = []
    for i in range(len(File)//2):
        E = float(File[i*2])
        eigenvector = [float(x) for x in File[i*2 + 1].split()]
        loaded.append([E, eigenvector])
    return loaded
        

